{%- extends 'base.html' %}
{% block title %}Error - {{ title }}{% endblock %}
{%- block body %}
    <div class="content">
        <p class="content">
        <article class="message is-danger">
            <div class="message-header">
                <p>{{ title }}</p>
            </div>
            <div class="message-body">
                {{ message }}
                {% if doc_link is defined %}
                    <p><a href="{{ doc_link }}">Click here for relevant documentation</a></p>
                {% endif %}
            </div>
        </article>
        </p>
    </div>
{% endblock -%}
